/*
 * drivers/mmc/card/aerial/aerial_drv.c
 *
 *  Copyright (C) 2008 Nissin Systems Co.,Ltd.
 *  Copyright (C) 2008-2009 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 2008-03-05    Created by Nissin Systems Co.,Ltd.
 * 2009-06-19   Modified by Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/kthread.h>
#include <linux/etherdevice.h>
#include <linux/leds.h>
#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/sdio.h>
#include <linux/mmc/sdio_func.h>
#include <linux/mmc/sdio_ids.h>

#include "aerial_drv.h"
#include "aerial_wid.h"
#include "aerial_ioctl.h"
#include "aerial_fw.h"
#include "aerial_sysfs.h"

struct sdio_bus_type {
	int id;
#define SDIO_DYNAMIC_WIDTH	(0)
#define SDIO_FORCE_1BIT		(1)
#define SDIO_FORCE_4BIT		(2)

	int now;
	int (*enable_wide)(struct mmc_card *card);
	int (*disable_wide)(struct mmc_card *card);
};

static struct sdio_bus_type bus_info = {
	.now = -1,
#if defined(ENABLE_FORCE_1BIT_MODE) && !defined(ENABLE_FORCE_4BIT_MODE)
	.id = SDIO_FORCE_1BIT,
#elif defined(ENABLE_FORCE_4BIT_MODE) && !defined(ENABLE_FORCE_1BIT_MODE)
	.id = SDIO_FORCE_4BIT,
#elif !defined(ENABLE_FORCE_4BIT_MODE) && !defined(ENABLE_FORCE_1BIT_MODE)
	.id = SDIO_DYNAMIC_WIDTH,
#else
#error "Unknown bus-width"
#endif
};

int aerial_log_flag = LOGLVL_1 | LOGLVL_2 ;

int
aerial_print(int level, const char *format, ...)
{
	if (aerial_log_flag & level) {
		int ret;
		va_list args;
		va_start(args, format);
		ret = vprintk(format, args);
		va_end(args);
		return ret;
	}
	return 0;
}

int
aerial_dump(int level, void *buf, int len)
{
	unsigned char *ptr = buf;
	int i;

	if (aerial_log_flag & level) {
		for (i=0; i<len; i++) {
			printk("%02x ", ptr[i]);
			if (i%16 == 15)
				printk("\n");
		}
		if (i%16 != 0)
			printk("\n");
	}
	return 0;
}

static int
aerial_sdio_f0_readb(struct sdio_func *func, unsigned int addr,
		     int *err_ret)
{
	int __num = func->num;
	unsigned char val;
	func->num = 0;
	val = sdio_readb(func, addr, err_ret);
	func->num = __num;
	return val;
}

static void
aerial_sdio_f0_writeb(struct sdio_func *func, unsigned char b,
		      unsigned int addr, int *err_ret)
{
	int __num = func->num;
	func->num = 0;
	sdio_writeb(func, b, addr, err_ret);
	func->num = __num;
}

static int
sdio_card_reset(struct mmc_card *card)
{
	struct sdio_func *func = card->sdio_func[0];
	unsigned char val;
	int ret;

	val = aerial_sdio_f0_readb(func, SDIO_CCCR_ABORT, &ret);
	if (ret)
		return ret;
	val |= 0x08;
	aerial_sdio_f0_writeb(func, val, SDIO_CCCR_ABORT, &ret);
	if (ret != -ETIMEDOUT)
		return ret;

	return 0;
}

struct aerial_restart_data {
	struct mmc_host *host;
	struct timer_list *timer;
	int state;
};
static struct aerial_restart_data restart_data;
static struct timer_list restart_timer;

static void
aerial_restart_timer_func(unsigned long _data)
{
	struct aerial_restart_data *data = (struct aerial_restart_data *)_data;
	struct mmc_host *host = data->host;
	struct timer_list *timer = data->timer;

	mmc_detect_change(host, msecs_to_jiffies(0));

	if (data->state++ == 0)
		mod_timer(timer, jiffies + msecs_to_jiffies(100));
}

int
aerial_restart(struct aerial_private *priv)
{
	struct sdio_func *func = priv->func;
	struct mmc_card *card = func->card;
	struct mmc_host *host = card->host;

	if (atomic_read(&priv->state) == AERIAL_STATE_NO_FW)
		return 0;

	sdio_claim_host(func);
	sdio_card_reset(card);
	sdio_release_host(func);

	restart_data.host = host;
	restart_data.timer = &restart_timer;
	restart_data.state = 0;
	setup_timer(restart_data.timer, aerial_restart_timer_func,
		    (unsigned long)&restart_data);
	mod_timer(restart_data.timer, jiffies + msecs_to_jiffies(1000));

	return 0;
}

static int
aerial_enable_wide(struct mmc_card *card)
{
	if (bus_info.now != 4) {
		sdio_enable_wide(card);
		bus_info.now = 4;
	}
	return 0;
}

static int
aerial_disable_wide(struct mmc_card *card)
{
	if (bus_info.now != 1) {
		sdio_disable_wide(card);
		bus_info.now = 1;
	}
	return 0;
}

/* sdio-lock of this function must be taken. */
int
aerial_send_prepare(struct sdio_func *func)
{
	unsigned char val;
	int retry = 100;
	int ret = -EIO;

	/* Set block length as word(32bit) to SD2AHB register (LSB) */
	retry = 10;
	while (retry--) {
		sdio_writeb(func, AERIAL_WRITE_WORD_SIZE & 0xff,
			    AERIAL_F1REG_SD2AHB_BLOCKLEN_LSB, &ret);
		if (!ret) {
			val = sdio_readb(func,
					 AERIAL_F1REG_SD2AHB_BLOCKLEN_LSB,
					 &ret);
			if (val == (AERIAL_WRITE_WORD_SIZE & 0xff))
				break;
		}
	}
	if (ret) {
		aer_err("failed set SD2AHB register(LSB)\n");
		return -EIO;
	}

	/* Set block length as word(32bit) to SD2AHB register (MSB) */
	retry = 10;
	while (retry--) {
		sdio_writeb(func, (AERIAL_WRITE_WORD_SIZE >> 8) | 0x80,
			    AERIAL_F1REG_SD2AHB_BLOCKLEN_MSB, &ret);
		if (!ret) {
			val = sdio_readb(func,
					 AERIAL_F1REG_SD2AHB_BLOCKLEN_MSB,
					 &ret);
			if (val == ((AERIAL_WRITE_WORD_SIZE >> 8) | 0x80))
				break;
		}
	}
	if (ret) {
		aer_err("failed set SD2AHB register(MSB)\n");
		return -EIO;
	}

	return 0;
}

static int
aerial_get_rx_blocks_from_register(struct sdio_func *func)
{
	u8 val;
	int ret;
	int blks;

	val = sdio_readb(func, AERIAL_F1REG_AHB2SD_BLOCKLEN_LSB, &ret);
	if (ret) {
		aer_err("failed get rx_blks from register(LSB)\n");
		return -1;
	}
	blks = val;

	val = sdio_readb(func, AERIAL_F1REG_AHB2SD_BLOCKLEN_MSB, &ret);
	if (ret) {
		aer_err("failed get rx_blks from register(MSB)\n");
		return -1;
	}
	blks |= (val << 8);
	blks *= sizeof(int);
	blks /= AERIAL_CHUNK_SIZE;

	return blks;
}

static void
aerial_rx_task(struct sdio_func *func)
{
	struct aerial_private *priv = sdio_get_drvdata(func);
	char *buf = (char *)&priv->rxbuf;
	unsigned char val;
	int ret;
	int read_blks;
	unsigned short data_type;
	unsigned short data_size;
	struct sk_buff *skb;
	int get_length;
	int error = 1;

	aerial_led_trigger_event(priv, active, LED_FULL);

	/* Check Interrupt status */
	val = sdio_readb(func, AERIAL_F1REG_FUNC1_INT_STATUS, &ret);
	if (ret) {
		aer_err("failed get interrupt status\n");
		goto RX_END;
	}

	/* Interrupt Pending */
	sdio_writeb(func, 0x01, AERIAL_F1REG_FUNC1_INT_PENDING, &ret);
	if (ret) {
		aer_err("failed clear interrupt status\n");
		goto RX_END;
	}

	for (;;) {
		val = sdio_readb(func, AERIAL_F1REG_FUNC1_INT_STATUS, &ret);
		if (ret) {
			aer_err("failed get interrupt status[2]\n");
			goto RX_END;
		}
		if (!val)
			break;
		aer_info("interrupt pending?\n");
		msleep(10);
	}

	/* Get read size */
	read_blks = aerial_get_rx_blocks_from_register(func);
	if (read_blks < 0)
		goto RX_END;

	get_length = read_blks * AERIAL_CHUNK_SIZE;

	/* first block read */
	ret = sdio_readsb(func, buf,
			  AERIAL_F1READBUFFER_ADDR, get_length);
	if (ret) {
		aer_err("failed first block read\n");
		goto RX_END;
	}

	/* get the data type  */
	data_type = (unsigned short)((buf[1] & 0xF0) >> 4);

	/* get the data size  */
	data_size = (unsigned short)(((buf[1] & 0x0F) << 8) |
				     buf[0]);

	if (data_type == HTYPE_CONFIG_RES) {
		if (buf[2] == CTYPE_RESPONSE) {
			memcpy(&priv->wid_response,
			       buf, data_size + 2);

			complete(&priv->wid_complete);
			goto RX_EXIT;
		}
		if (buf[2] == CTYPE_INFO) {
			if (buf[9] == 0x0) {
				aer_info("disconnected!\n");
				netif_stop_queue(priv->netdev);
				netif_carrier_off(priv->netdev);
				aerial_led_trigger_event(priv, link, LED_OFF);
			} else {
				aer_info("connected!\n");
				netif_carrier_on(priv->netdev);
				netif_wake_queue(priv->netdev);
				aerial_led_trigger_event(priv, link, LED_FULL);
			}
			if (atomic_read(&priv->state) ==
			    AERIAL_STATE_NO_FW)
				complete(&priv->firmware_ready);

			goto RX_EXIT;
		}
		goto RX_EXIT;
	}

	/* Data IN packet */
	skb = dev_alloc_skb(data_size + 2 + NET_IP_ALIGN);
	if (!skb) {
		priv->net_stats.rx_dropped++;
		aer_err("failed skb memory allocate error\n");
		goto RX_END;
	}
	skb->dev = priv->netdev;
	skb_reserve(skb, NET_IP_ALIGN);
	skb_put(skb, data_size + 2);
	memcpy(skb->data, &buf[2], data_size);

	skb->protocol = eth_type_trans(skb, skb->dev);

	ret = netif_rx(skb);
	if (ret == NET_RX_DROP) {
		priv->net_stats.rx_dropped++;
		aer_info("packet dropped\n");
		dev_kfree_skb_any(skb);
	}
	priv->net_stats.rx_packets++;
	priv->net_stats.rx_bytes += data_size;

	error = 0;

RX_END:
	if (error)
		priv->net_stats.rx_errors++;

RX_EXIT:
	aerial_led_trigger_event(priv, active, LED_OFF);

	return;
}

static void
aerial_tx_task(struct sk_buff *skb, struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct aerial_packet txbuf;
	struct sdio_func *func = priv->func;
	unsigned long flags;
	int ret;

	if (atomic_read(&priv->state) != AERIAL_STATE_RUNNING) {
		aer_err("tx_task: driver state is not running!\n");
		dev_kfree_skb_any(skb);
		return;
	}

	txbuf.h.len = skb->len;
	txbuf.h.type = HTYPE_DATA_OUT;
	memcpy(txbuf.message, skb->data, skb->len);
	dev_kfree_skb_any(skb);

	dev->trans_start = jiffies;

	local_irq_save(flags);
	if (atomic_read(&priv->power_state) == AERIAL_PWR_STATE_SLEEP) {
		local_irq_restore(flags);
		return;
	}
	sdio_claim_host(func);
	local_irq_restore(flags);

	aerial_led_trigger_event(priv, active, LED_FULL);

	if (bus_info.enable_wide)
		bus_info.enable_wide(func->card);

	ret = aerial_send_prepare(func);
	if (ret) {
		aer_err("failed send prepare\n");
		priv->net_stats.tx_errors++;
		goto tx_end;
	}

	ret = sdio_writesb(func, AERIAL_F1WRITEBUFFER_ADDR, &txbuf,
			   AERIAL_WRITE_SIZE);
	if (ret) {
		aer_err("failed send data\n");
		priv->net_stats.tx_errors++;
		goto tx_end;
	}

	priv->net_stats.tx_packets++;
	priv->net_stats.tx_bytes += txbuf.h.len;

tx_end:
	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);
	sdio_release_host(func);

	aerial_led_trigger_event(priv, active, LED_OFF);

	return;
}

static void
aerial_irq(struct sdio_func *func)
{
	if (bus_info.enable_wide)
		bus_info.enable_wide(func->card);

	aerial_rx_task(func);

	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);

	return;
}

static int
aerial_open(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	if (atomic_read(&priv->state) == AERIAL_STATE_NO_FW) {
		aer_err("firmware was not loadding.\n");
		return -EIO;
	}

	atomic_set(&priv->state, AERIAL_STATE_RUNNING);
	netif_start_queue(dev);

	return 0;
}

static int
aerial_close(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	atomic_set(&priv->state, AERIAL_STATE_READY);
	netif_stop_queue(dev);

	return 0;
}

static int
aerial_hardstart(struct sk_buff *skb, struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct aerial_task_info *task = &priv->tx_task;
	struct tx_args *args;
	unsigned long flags;

	args = kzalloc(sizeof(struct tx_args), GFP_KERNEL);
	if (args == NULL) {
		dev_kfree_skb_any(skb);
		aer_err("failed tx_args memory allocation\n");
		return -ENOMEM;
	}

	args->dev = dev;
	args->skb = skb;

	spin_lock_irqsave(&task->lock, flags);
	list_add_tail(&args->entry, &task->tx_args_list);
	spin_unlock_irqrestore(&task->lock, flags);

	schedule_delayed_work(&task->queue, usecs_to_jiffies(0));

	return 0;
}

static int
aerial_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	return 0;
}

static struct net_device_stats *
aerial_getstats(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	return &priv->net_stats;
}

static void
aerial_do_task(struct work_struct *work)
{
	struct aerial_task_info *task = container_of(work,
						     struct aerial_task_info,
						     queue.work);
	struct aerial_private *priv = container_of(task,
						   struct aerial_private,
						   tx_task);

	complete(&priv->tx_thread_wait);
}

static int
aerial_tx_thread(void *_priv)
{
	struct aerial_private *priv = _priv;
	struct aerial_task_info *task = &priv->tx_task;
	struct tx_args *args;
	unsigned long flags;

	init_completion(&priv->tx_thread_wait);

	do {
		int timeout;

		timeout = wait_for_completion_interruptible_timeout(
						&priv->tx_thread_wait,
						msecs_to_jiffies(100));
		set_current_state(TASK_RUNNING);
		if (timeout > 0) {
			spin_lock_irqsave(&task->lock, flags);
			while (!list_empty(&task->tx_args_list)) {
				args = list_entry(task->tx_args_list.next,
						  struct tx_args, entry);
				list_del(&args->entry);
				spin_unlock_irqrestore(&task->lock, flags);

				aerial_tx_task(args->skb, args->dev);
				kfree(args);

				spin_lock_irqsave(&task->lock, flags);
			}
			spin_unlock_irqrestore(&task->lock, flags);
		}
	} while (!kthread_should_stop());

	return 0;
}

static int
aerial_task_init(struct aerial_private *priv)
{
	struct aerial_task_info *task = &priv->tx_task;

	memset(task, 0, sizeof(struct aerial_task_info));

	spin_lock_init(&task->lock);
	INIT_LIST_HEAD(&task->tx_args_list);
	INIT_DELAYED_WORK(&task->queue, aerial_do_task);

	priv->tx_thread = kthread_run(aerial_tx_thread, priv, "%s.tx/%s",
				      AERIAL_DRVINFO, priv->netdev->name);
	if (IS_ERR(priv->tx_thread)) {
		aer_err("Could not create thread.\n");
		return PTR_ERR(priv->tx_thread);
	}

	return 0;
}

static void
aerial_task_exit(struct aerial_private *priv)
{
	if (priv->tx_thread) {
		kthread_stop(priv->tx_thread);
		priv->tx_thread = NULL;
	}
}

static int
aerial_sdio_if_init(struct sdio_func *func)
{
	unsigned char val;
	int ret;

	sdio_claim_host(func);

	ret = sdio_enable_func(func);
	if (ret) {
		aer_err("failed enable func\n");
		goto init_end;
	}

	val = sdio_readb(func, AERIAL_F1REG_INT_MASK_REG, &ret);
	if (ret) {
		aer_err("failed read f1:int-mask\n");
		goto init_end;
	}
	if (!val) {
		sdio_writeb(func, 0x01, AERIAL_F1REG_INT_MASK_REG, &ret);
		if (ret) {
			aer_err("failed set f1:int-mask\n");
			goto init_end;
		}
	}

	sdio_claim_irq(func, aerial_irq);

	/* Set block length */
	ret = sdio_set_block_size(func, AERIAL_CHUNK_SIZE);
	if (ret) {
		aer_err("failed set block size\n");
		goto init_end;
	}

	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);

init_end:
	sdio_release_host(func);

	return ret;
}

static int
sdio_aerial_probe(struct sdio_func *func, const struct sdio_device_id *id)
{
	struct aerial_private *priv;
	struct net_device *dev;
	int ret;

	/* net device initialize */
	dev = alloc_netdev(sizeof(struct aerial_private),
			   "wlan%d", ether_setup);
	if (!dev) {
		aer_err("could not allocate device\n");
		return -ENOMEM;
	}

	dev->open = aerial_open;
	dev->stop = aerial_close;
	dev->hard_start_xmit = aerial_hardstart;
	dev->do_ioctl = aerial_ioctl;
	dev->get_stats = aerial_getstats;

	aerial_wireless_attach(dev);

	/* card specific capability */
	func->card->ext_caps = MMC_CARD_CAPS_FORCE_BLKXMIT;
	bus_info.now = -1;
	switch (bus_info.id) {
	case SDIO_DYNAMIC_WIDTH:
		bus_info.enable_wide = aerial_enable_wide;
		bus_info.disable_wide = aerial_disable_wide;
		break;
	case SDIO_FORCE_1BIT:
		bus_info.enable_wide = NULL;
		bus_info.disable_wide = aerial_disable_wide;
		break;
	case SDIO_FORCE_4BIT:
		func->card->ext_caps |= MMC_CARD_CAPS_FORCE_CLK_KEEP;
		bus_info.enable_wide = aerial_enable_wide;
		bus_info.disable_wide = NULL;
		break;
	}

	/* Aerial initialize */
	priv = dev->priv;
	priv->netdev = dev;
	priv->func = func;

	sdio_set_drvdata(func, priv);

	atomic_set(&priv->state, AERIAL_STATE_NO_FW);
	atomic_set(&priv->power_state, AERIAL_PWR_STATE_ACTIVE);
	sema_init(&priv->wid_req_lock, 1);
	init_completion(&priv->wid_complete);
	init_completion(&priv->firmware_ready);

	priv->fw_image = kmalloc(FIRMWARE_SIZE_MAX, GFP_KERNEL);
	priv->fw_size  = 0;

	ret = aerial_sdio_if_init(func);
	if (ret)
		goto if_err;

	ret = register_netdev(dev);
	if (ret) {
		aer_err("failed register netdev\n");
		goto netdev_err;
	}

	ret = aerial_sysfs_init(priv);
	if (ret) {
		aer_err("failed register sysfs-entry\n");
		goto sysfs_err;
	}

	ret = aerial_task_init(priv);
	if (ret)
		goto task_err;

	aerial_led_trigger_register(priv, active);
	aerial_led_trigger_register(priv, link);
	aerial_led_trigger_event(priv, active, LED_OFF);
	aerial_led_trigger_event(priv, link, LED_OFF);

	printk(KERN_INFO "%s: registerd \"%s\" device as %s\n",
	       mmc_hostname(func->card->host), AERIAL_DRVINFO, dev->name);

	return 0;

task_err:
	aerial_sysfs_remove(priv);
sysfs_err:
	unregister_netdev(dev);
netdev_err:
if_err:
	free_netdev(dev);

	return ret;
}

static void
sdio_aerial_remove(struct sdio_func *func)
{
	struct aerial_private *priv = sdio_get_drvdata(func);

	aerial_led_trigger_unregister(priv, active);
	aerial_led_trigger_unregister(priv, link);

	aerial_task_exit(priv);

	sdio_claim_host(func);
	sdio_release_irq(func);
	sdio_disable_func(func);
	sdio_release_host(func);

	unregister_netdev(priv->netdev);
	aerial_sysfs_remove(priv);

	kfree(priv->fw_image);
	free_netdev(priv->netdev);

	return;
}

static const struct sdio_device_id sdio_aerial_ids[] = {
	{ SDIO_CLASS_WLAN, SDIO_VENDOR_ID_ROHM, SDIO_DEVICE_ID_ROHM_BW9419 },
	{ /* end: all zeros */ },
};

MODULE_DEVICE_TABLE(sdio, sdio_aerial_ids);

static struct sdio_driver sdio_aerial_driver = {
	.probe = sdio_aerial_probe,
	.remove = sdio_aerial_remove,
	.name = "sdio_aerial",
	.id_table = sdio_aerial_ids,
};

static int __init
sdio_aerial_init(void)
{
	int ret;

	ret = sdio_register_driver(&sdio_aerial_driver);
	if (ret)
		aer_err("Registration failure.(%d)\n", ret);
	else
		aer_info("Version %s Load.\n", AERIAL_VERSION);

	return ret;
}

static void __exit
sdio_aerial_exit(void)
{
	sdio_unregister_driver(&sdio_aerial_driver);
	aer_info("Version %s unloaded.\n", AERIAL_VERSION);
}

module_init(sdio_aerial_init);
module_exit(sdio_aerial_exit);

MODULE_AUTHOR("Nissin Systems Co.,Ltd.");
MODULE_LICENSE("GPL");
