/*
 * drivers/mmc/card/aerial/aerial_ioctl.c
 *
 *  Copyright (C) 2008 Nissin Systems Co.,Ltd.
 *  Copyright (C) 2008-2009 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 2008-03-05    Created by Nissin Systems Co.,Ltd.
 * 2009-06-19   Modified by Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/mmc/core.h>
#include <linux/mmc/sdio_func.h>
#include <linux/mmc/sdio_ids.h>
#include <linux/mmc/sdio.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <net/iw_handler.h>
#include "aerial_drv.h"
#include "aerial_ioctl.h"
#include "aerial_fw.h"
#include "aerial_wid.h"

#define CHARCODE_TO_NUMBIN(cc, ret)		\
({						\
	unsigned char num;			\
	ret = 0;				\
	switch ((cc)) {				\
	case '0' ... '9':			\
		num = (cc) - '0'; break;	\
	case 'a' ... 'f':			\
		num = (cc) - 'a' + 10; break;	\
	case 'A' ... 'F':			\
		num = (cc) - 'A' + 10; break;	\
	default:				\
		num = 0; ret = -1; break;	\
	}					\
	num;					\
})

#define NUMBIN_TO_CHARCODE(num)			\
({						\
	char cc;				\
	switch ((num)) {			\
	default:				\
		cc = 0;				\
		/* FALL THROUGH */		\
	case 0 ... 9:				\
		cc = (num) + '0'; break;	\
	case 10 ... 15:				\
		cc = (num) - 10 + 'A'; break;	\
	}					\
	cc;					\
})

static long aerial_support_freq[] = {
	/* 0: no support */
	[ 0] =    0, [ 1] = 2412, [ 2] = 2417, [ 3] = 2422,
	[ 4] = 2427, [ 5] = 2432, [ 6] = 2437, [ 7] = 2442,
	[ 8] = 2447, [ 9] = 2452, [10] = 2457, [11] = 2462,
	[12] = 2467, [13] = 2472, [14] = 2484,
};

static int aerial_rate_table[] = {
	[ 1] =  1000000, [ 2] =  2000000, [ 5] =  5500000, [ 6] =  6000000,
	[ 9] =  9000000, [11] = 11000000, [12] = 12000000, [18] = 18000000,
	[24] = 24000000, [36] = 36000000, [48] = 48000000, [54] = 54000000,
};

static unsigned char aerial_support_rate[][16] = {
	/* reserved     */ {},
	/* HighSpeed    */ { 6,  9, 12, 18, 24, 36, 48, 54},
	/* Compatible 1 */ { 1,  2,  5,  6,  9, 11, 12, 18, 24, 36, 48, 54},
	/* Compatible 2 */ { 1,  2,  5,  6,  9, 11, 12, 18, 24, 36, 48, 54},
};

static inline long
_pow(long m, short e)
{
	long r = m;
	int i;
	if (e == 0)
		return 1;
	if (e < 0)
		/* not impremented */
		return -1;
	for (i=1; i<e; i++)
		r *= m;

	return r;
}

static int
support_channel(unsigned char ch)
{
	if ((ch <= ARRAY_SIZE(aerial_support_freq))
	    && (aerial_support_freq[ch]))
		return 1;

	return 0;
}

static unsigned char
freq_to_channel(struct iw_freq *freq)
{
	long f;
	int i;

	if (!freq || freq->m == 0)
		return 0xff;

	f = freq->m * _pow(10, freq->e) / _pow(10, 6);
	for (i=0; i<ARRAY_SIZE(aerial_support_freq); i++)
		if (aerial_support_freq[i] == f)
			return i;

	return 0xff;
}

static int
support_rate(unsigned char mode, unsigned char rate)
{
	int i;

	for (i=0; i<ARRAY_SIZE(aerial_support_rate[mode]); i++)
		if (aerial_support_rate[mode][i] == rate)
			return 1;

	return 0;
}

static unsigned char
value_to_rate(int value)
{
	int i;

	for (i=0; i<ARRAY_SIZE(aerial_rate_table); i++)
		if (aerial_rate_table[i] == value)
			return i;

	return 0xff;
}

static int
aerial_iw_not_support(struct net_device *dev, struct iw_request_info *info,
		      void *arg, char *extra)
{
	return -EOPNOTSUPP;
}

static int
aerial_iw_giwname(struct net_device *dev, struct iw_request_info *info,
		  char *name, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);

	if (atomic_read(&priv->power_state) == AERIAL_PWR_STATE_SLEEP)
		strncpy(name, "sleeping", IFNAMSIZ);
	else
		strncpy(name, "IEEE 802.11bg", IFNAMSIZ);

	return 0;
}

static int
aerial_iw_siwfreq(struct net_device *dev, struct iw_request_info *info,
		  struct iw_freq *freq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char ch;
	int ret;

	if (freq->e == 0)
		ch = freq->m;
	else
		ch = freq_to_channel(freq);

	if (!support_channel(ch))
		return -EINVAL;

	ret = aerial_set_channel(priv, ch);
	if (ret)
		return ret;

	return 0;
}

static int
aerial_iw_giwfreq(struct net_device *dev, struct iw_request_info *info,
		  struct iw_freq *freq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char ch;
	int ret;

	ret = aerial_get_channel(priv, &ch);
	if (ret)
		return ret;

	if (!support_channel(ch))
		return -EINVAL;

	freq->e = 0;
	freq->m = ch;

	return 0;
}

static int
aerial_iw_siwmode(struct net_device *dev, struct iw_request_info *info,
		  __u32 *mode, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char type;
	int ret;

	switch (*mode) {
	case IW_MODE_INFRA:
		type = AERIAL_BSS_INFRA; break;
	case IW_MODE_ADHOC:
		type = AERIAL_BSS_ADHOC; break;
	default:
		return -EOPNOTSUPP;
	}

	ret = aerial_set_bsstype(priv, type);
	if (ret)
		return ret;

	return 0;
}

static int
aerial_iw_giwmode(struct net_device *dev, struct iw_request_info *info,
		  __u32 *mode, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char type;
	int ret;

	ret = aerial_get_bsstype(priv, &type);
	if (ret)
		return ret;

	switch (type) {
	case AERIAL_BSS_INFRA:
		*mode = IW_MODE_INFRA; break;
	case AERIAL_BSS_ADHOC:
		*mode = IW_MODE_ADHOC; break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int
aerial_iw_giwrange(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *data, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct iw_range *range = (struct iw_range *)extra;
	unsigned char rate_mode;
	int ret;
	int i;

	memset(range, 0, sizeof(struct iw_range));
	data->length = sizeof(struct iw_range);

	ret = aerial_get_ratemode(priv, &rate_mode);
	if (ret)
		return ret;

	range->num_channels = 0;
	for (i=0; i<ARRAY_SIZE(aerial_support_freq); i++) {
		if (aerial_support_freq[i]) {
			range->freq[range->num_channels].i = i;
			range->freq[range->num_channels].m =
				aerial_support_freq[i];
			range->freq[range->num_channels].e = 6;
			range->num_channels++;
		}
	}
	range->num_frequency = range->num_channels;

	range->num_bitrates = 0;
	for (i=0; i<ARRAY_SIZE(aerial_support_rate[rate_mode]); i++) {
		unsigned char id = aerial_support_rate[rate_mode][i];
		if (id) {
			range->bitrate[range->num_bitrates] =
				aerial_rate_table[id];
			range->num_bitrates++;
		}
	}

	range->max_encoding_tokens = AERIAL_KEYINDEX_MAX;
	range->num_encoding_sizes = 2;
	range->encoding_size[0] = 5; /* 64bits WEP */
	range->encoding_size[1] = 13; /* 128bits WEP */
	range->encoding_size[2] = 32; /* 256bits WPA-PSK */

	range->num_txpower = 0;
	range->txpower_capa = IW_TXPOW_DBM;
	range->min_retry = 0;
	range->max_retry = 0;
	range->max_qual.level = 0;
	range->max_qual.noise = 0;
	range->sensitivity = 0;

	range->min_rts = AERIAL_RTS_MIN;
	range->max_rts = AERIAL_RTS_MAX;
	range->min_frag = AERIAL_FRAG_MIN;
	range->max_frag = AERIAL_FRAG_MAX;
	range->enc_capa = (IW_ENC_CAPA_WPA |
			   IW_ENC_CAPA_WPA2 |
			   IW_ENC_CAPA_CIPHER_TKIP |
			   IW_ENC_CAPA_CIPHER_CCMP);

	range->scan_capa = IW_SCAN_CAPA_ESSID;
	range->event_capa[0] = (IW_EVENT_CAPA_K_0 |
				IW_EVENT_CAPA_MASK(SIOCGIWAP) |
				IW_EVENT_CAPA_MASK(SIOCGIWSCAN));
	range->event_capa[1] = IW_EVENT_CAPA_K_1;

	range->we_version_source = SUPPORTED_WIRELESS_EXT;
	range->we_version_compiled = WIRELESS_EXT;

	return 0;
}

static int
aerial_iw_siwap(struct net_device *dev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	return -EOPNOTSUPP;
}

static int
aerial_iw_giwap(struct net_device *dev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char bssid[6];
	int ret;

	ret = aerial_get_bssid(priv, bssid, 6);
	if (ret)
		return ret;

	memcpy(&awrq->sa_data, bssid, 6);
	awrq->sa_family = ARPHRD_ETHER;

	return 0;
}

static int
aerial_iw_siwscan(struct net_device *dev, struct iw_request_info *info,
		  struct iw_point *srq, char *extra)
{
	return 0;
}

static int
aerial_iw_giwscan(struct net_device *dev, struct iw_request_info *info,
		  struct iw_point *srq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct aerial_survey_res survey[2];
	struct iw_event iwe;
	char *cev = extra, *cev_end = extra + IW_SCAN_MAX_DATA;
	unsigned char old_scantype, old_scanfilter;
	unsigned char old_ch;
	int ret;
	int i, j;

	ret = aerial_get_scan_type(priv, &old_scantype);
	if (ret)
		return ret;
	aerial_get_scan_filter(priv, &old_scanfilter);
	aerial_get_channel(priv, &old_ch);

	aerial_set_scan_type(priv, 0x0);
	aerial_set_scan_filter(priv, 0x0);

	aerial_set_sitesv(priv, 0x01);
	aerial_set_startscan(priv, 0x01);
	for (i=0; i<10; i++) {
		unsigned char val;
		aerial_get_startscan(priv, &val);
		if (!val)
			break;
		msleep(10);
	}

	for (i=0; i<2; i++) {
		int size = sizeof(struct aerial_survey_res);
		memset(&survey[i], 0, size);
		ret = aerial_get_site_survey_results(priv, &survey[i], size);
		if (ret)
			return -EIO;
	}

	aerial_set_sitesv(priv, 0x02);

	aerial_set_scan_type(priv, old_scantype);
	aerial_set_scan_filter(priv, old_scanfilter);
	aerial_set_channel(priv, old_ch);

	for (i=0; i<2; i++) {
		int nr_info = (survey[i].size /
			       sizeof(struct aerial_survey_info));
		for (j=0; j<nr_info; j++) {
			struct aerial_survey_info *info;
			info = &survey[i].info[j];

			/* First entry *MUST* be the BSSID */
			iwe.cmd = SIOCGIWAP;
			iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
			memcpy(iwe.u.ap_addr.sa_data, info->bssid, ETH_ALEN);
			cev = iwe_stream_add_event(cev, cev_end, &iwe,
						   IW_EV_ADDR_LEN);

			/* SSID */
			iwe.cmd = SIOCGIWESSID;
			iwe.u.data.length = strlen(info->ssid);
			iwe.u.data.flags = 1;
			cev = iwe_stream_add_point(cev, cev_end, &iwe,
						   info->ssid);

			/* Mode */
			iwe.cmd = SIOCGIWMODE;
			if (info->bsstype & 1)
				iwe.u.mode = IW_MODE_ADHOC;
			else
				iwe.u.mode = IW_MODE_MASTER;
			cev = iwe_stream_add_event(cev, cev_end, &iwe,
						   IW_EV_UINT_LEN);

			/* Frequency */
			iwe.cmd = SIOCGIWFREQ;
			iwe.u.freq.m = aerial_support_freq[info->channel];
			iwe.u.freq.e = 6;
			cev = iwe_stream_add_event(cev, cev_end, &iwe,
						   IW_EV_FREQ_LEN);

			/* Encryption capability */
			iwe.cmd = SIOCGIWENCODE;
			if (info->security & AERIAL_CRYPT_ENABLE)
				iwe.u.data.flags = (IW_ENCODE_ENABLED |
						    IW_ENCODE_NOKEY);
			else
				iwe.u.data.flags = IW_ENCODE_DISABLED;

			iwe.u.data.length = 0;
			cev = iwe_stream_add_point(cev, cev_end, &iwe, NULL);
		}
	}

	srq->length = cev - extra;
	srq->flags = 0;

	return 0;
}

static int
aerial_iw_siwessid(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *data, char *essid)
{
	struct aerial_private *priv = netdev_priv(dev);

	return aerial_set_ssid(priv, essid, data->length);
}

static int
aerial_iw_giwessid(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *data, char *essid)
{
	struct aerial_private *priv = netdev_priv(dev);
	int size = min(AERIAL_SSID_MAX, (int)data->length);

	memset(essid, 0, size);
	if (aerial_get_ssid(priv, essid, size) != 0)
		return -EINVAL;

	data->flags = 1;
	data->length = strlen(essid);

	return 0;
}

static int
aerial_iw_siwrate(struct net_device *dev, struct iw_request_info *info,
		  struct iw_param *rrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char ratemode, rate;
	int ret;

	if (rrq->fixed == 0) {
		/* Set Auto */
		rate = 0;
	} else {
		ret = aerial_get_ratemode(priv, &ratemode);
		if (ret)
			return ret;

		rate = value_to_rate(rrq->value);
		if (!support_rate(ratemode, rate))
			return -EINVAL;
	}

	ret = aerial_set_rate(priv, rate);

	return ret;
}

static int
aerial_iw_giwrate(struct net_device *dev, struct iw_request_info *info,
		  struct iw_param *rrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char rate;
	int ret;

	ret = aerial_get_rate(priv, &rate);
	if (ret)
		return ret;

	rrq->value = aerial_rate_table[rate];
	rrq->fixed = (rrq->value ? 1 : 0);

	return 0;
}

static int
aerial_iw_siwenc(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *dwrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char crypt = AERIAL_CRYPT_ENABLE | AERIAL_CRYPT_WEP;
	unsigned char buf[32];
	int index;
	int ret;
	int i;

	if (dwrq->flags & IW_ENCODE_DISABLED) {
		return aerial_set_cryptmode(priv, AERIAL_CRYPT_DISABLE);
	}

	if (dwrq->flags & IW_ENCODE_NOKEY)
		return -EINVAL;

	index = dwrq->flags & IW_ENCODE_INDEX;
	if ((index < 0) || (AERIAL_KEYINDEX_MAX < index))
		return -EINVAL;
	if (index != 0)
		index--;

	if (!extra)
		return -EINVAL;
	switch (dwrq->length) {
	case (AERIAL_WEPKEY_MAX / 2):
		crypt |= AERIAL_CRYPT_WEPSZ;
		/* FALL THROUGH */
	case (AERIAL_WEPKEY_MIN / 2):
		break;
	default:
		return -EINVAL;
	}

	memset(buf, 0, sizeof(buf));
	for (i=0; i<dwrq->length; i++) {
		buf[i*2  ] = NUMBIN_TO_CHARCODE((extra[i] >> 4) & 0x0f);
		buf[i*2+1] = NUMBIN_TO_CHARCODE(extra[i] & 0x0f);
	}

	ret = aerial_set_cryptmode(priv, crypt);
	if (ret)
		return ret;
	ret = aerial_set_wepkey(priv, index, buf, strlen(buf));
	if (ret)
		return ret;

	return 0;
}

static int
aerial_iw_giwenc(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *dwrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char crypt;
	unsigned char enckey[32];
	int encsize;
	int index;
	int ret;
	int i;

	ret = aerial_get_cryptmode(priv, &crypt);
	if (ret)
		return ret;

	index = dwrq->flags & IW_ENCODE_INDEX;
	if ((index < 0) || (AERIAL_KEYINDEX_MAX < index))
		return -EINVAL;
	if (index != 0)
		index--;

	switch (crypt) {
	case AERIAL_CRYPT_WEP64:
		encsize = 10;
		break;
	case AERIAL_CRYPT_WEP128:
		encsize = 26;
		break;
	default:
		dwrq->length = 0;
		dwrq->flags = IW_ENCODE_DISABLED;
		return 0;
	}

	memset(enckey, 0, 32);
	ret = aerial_get_wepkey(priv, index, enckey, 32);
	if (ret)
		return ret;

	dwrq->length = encsize / 2;
	dwrq->flags = (index + 1) | IW_ENCODE_ENABLED;
	for (i=0; i<dwrq->length; i++) {
		extra[i] = (char)CHARCODE_TO_NUMBIN(enckey[i*2+1], ret);
		extra[i] |= (char)(CHARCODE_TO_NUMBIN(enckey[i*2], ret) << 4);
	}

	return 0;
}

static int
aerial_iw_siwpower(struct net_device *dev, struct iw_request_info *info,
		   struct iw_param *wrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char mode;

	if (wrq->disabled) {
		mode = AERIAL_PMMODE_ACT;
	} else {
		switch (wrq->flags & IW_POWER_MODE) {
		case IW_POWER_ON:
			mode = AERIAL_PMMODE_SAVE;
			break;
		default:
			return -EOPNOTSUPP;
		}
	}

	return aerial_set_powerman(priv, mode);
}

static int
aerial_iw_giwpower(struct net_device *dev, struct iw_request_info *info,
		   struct iw_param *rrq, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char val;
	int ret;

	ret = aerial_get_powerman(priv, &val);
	if (ret)
		return ret;

	if (val == AERIAL_PMMODE_ACT)
		rrq->disabled = 1;
	else
		rrq->disabled = 0;

	return 0;
}

static int
aerial_iw_siwauth(struct net_device *dev, struct iw_request_info *info,
		  struct iw_param *dwrq, char *extra)
{
	return 0;
}

static int
aerial_iw_siwencodeext(struct net_device *dev, struct iw_request_info *info,
		       struct iw_point *dwrq, char *extra)
{
	return -EOPNOTSUPP;
}

static int
aerial_ip_fwload(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);

	return aerial_firmware_setup(priv);
}

static int
aerial_ip_setmac(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *data, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	char *str = data->pointer;
	unsigned char mac[6];
	int i;

	if (data->length != 12)
		return -EINVAL;

	for (i=0; i<12; i++) {
		unsigned char num;
		int ret;
		num = CHARCODE_TO_NUMBIN(str[i], ret);
		if (ret)
			return -EINVAL;

		if (i%2)
			mac[i/2] |= num;
		else
			mac[i/2] = num << 4;
	}

	return aerial_set_macaddr(priv, mac, 6);
}

static int
aerial_ip_getmac(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *data, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char mac[6];
	int ret;

	ret = aerial_get_macaddr(priv, mac, 6);
	if (ret)
		return ret;

	sprintf(extra, "%02x%02x%02x%02x%02x%02x",
		mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

	data->length = strlen(extra);

	return 0;
}

struct aerial_crypt_mode {
	char *label;
	unsigned char val;
};
struct aerial_crypt_mode crypts[] = {
	{ "NONE", 0x00, },
	{ "WEP64", 0x03, },
	{ "WEP128", 0x07, },
	{ "WPA-AES", 0x29 },
	{ "WPA-CCMP", 0x29 },
	{ "WPA-TKIP", 0x49, },
	{ "WPA2-AES", 0x31, },
	{ "WPA2-CCMP", 0x31, },
	{ "WPA2-TKIP", 0x51, },
};

static int
aerial_ip_setcrypt(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	int i;

	for (i=0; i<ARRAY_SIZE(crypts); i++)
		if (strcasecmp(crypts[i].label, extra) == 0)
			return aerial_set_cryptmode(priv, crypts[i].val);

	return -EINVAL;
}

static int
aerial_ip_getcrypt(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char val;
	int ret;
	int i;

	ret = aerial_get_cryptmode(priv, &val);
	if (ret)
		return ret;

	for (i=0; i<ARRAY_SIZE(crypts); i++) {
		if (crypts[i].val == val) {
			sprintf(extra, "%s\n", crypts[i].label);
			wri->length = strlen(extra);
			return 0;
		}
	}

	sprintf(extra, "unknown(%02x)\n", val);
	wri->length = strlen(extra);
	return 0;
}

static int
aerial_ip_setpsk(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);

	return aerial_set_passps(priv, extra, wri->length - 1);
}

static int
aerial_ip_getpsk(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	unsigned char pskpass[AERIAL_PSKPASS_MAX + 1];
	int ret;

	ret = aerial_get_passps(priv, pskpass, sizeof(pskpass));
	if (ret)
		return ret;

	strcpy(extra, pskpass);
	wri->length = strlen(extra);

	return 0;
}

static int
aerial_ip_setpsave(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	int *param = (int *)extra;

	return aerial_set_powerman(priv, param[0]);
}

static int
aerial_ip_getpsave(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	int *param = (int *)extra;
	unsigned char val;
	int ret;

	ret = aerial_get_powerman(priv, &val);
	if (ret)
		return ret;

	param[0] = (int)val;

	return 0;
}

static int
aerial_ip_setsleep(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	int *param = (int *)extra;

	return aerial_set_psctl(priv, param[0]);
}

static int
aerial_ip_getsleep(struct net_device *dev, struct iw_request_info *info,
		   struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);
	int *param = (int *)extra;

	param[0] = (int)atomic_read(&priv->power_state);

	return 0;
}

static int
aerial_ip_reset(struct net_device *dev, struct iw_request_info *info,
		struct iw_point *wri, char *extra)
{
	struct aerial_private *priv = netdev_priv(dev);

	return aerial_restart(priv);
}

static int
aerial_ip_setlog(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *wri, char *extra)
{
	int *param = (int *)extra;
	int val = param[0];

	aerial_log_flag = val;

	return 0;
}

static int
aerial_ip_getlog(struct net_device *dev, struct iw_request_info *info,
		 struct iw_point *wri, char *extra)
{
	int *param = (int *)extra;

	param[0] = aerial_log_flag;

	return 0;
}

static struct iw_statistics *
aerial_iw_getstats(struct net_device *dev)
{
	return NULL;
}

#define IW_IOCTL(x, func) [(x) - SIOCSIWCOMMIT] = (iw_handler)func
static const iw_handler aerial_iw_handlers[] = {
	/* Wireless Identification */
	IW_IOCTL(SIOCSIWCOMMIT,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWNAME,		aerial_iw_giwname),
	/* Basic operations */
	IW_IOCTL(SIOCSIWNWID,		NULL),
	IW_IOCTL(SIOCGIWNWID,		aerial_iw_not_support),
	IW_IOCTL(SIOCSIWFREQ,		aerial_iw_siwfreq),
	IW_IOCTL(SIOCGIWFREQ,		aerial_iw_giwfreq),
	IW_IOCTL(SIOCSIWMODE,		aerial_iw_siwmode),
	IW_IOCTL(SIOCGIWMODE,		aerial_iw_giwmode),
	IW_IOCTL(SIOCSIWSENS,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWSENS,		aerial_iw_not_support),
	/* Informative stuff */
	IW_IOCTL(SIOCSIWRANGE,		NULL),
	IW_IOCTL(SIOCGIWRANGE,		aerial_iw_giwrange),
	IW_IOCTL(SIOCSIWPRIV,		NULL),
	IW_IOCTL(SIOCGIWPRIV,		NULL),
	IW_IOCTL(SIOCSIWSTATS,		NULL),
	IW_IOCTL(SIOCGIWSTATS,		NULL),
	/* Spy support */
	IW_IOCTL(SIOCSIWSPY,		NULL),
	IW_IOCTL(SIOCGIWSPY,		NULL),
	IW_IOCTL(SIOCSIWTHRSPY,		NULL),
	IW_IOCTL(SIOCGIWTHRSPY,		NULL),
	/* Access Point manipulation */
	IW_IOCTL(SIOCSIWAP,		aerial_iw_siwap),
	IW_IOCTL(SIOCGIWAP,		aerial_iw_giwap),
	IW_IOCTL(SIOCGIWAPLIST,		NULL),
	IW_IOCTL(SIOCSIWSCAN,		aerial_iw_siwscan),
	IW_IOCTL(SIOCGIWSCAN,		aerial_iw_giwscan),
	/* 802.11 specific support */
	IW_IOCTL(SIOCSIWESSID,		aerial_iw_siwessid),
	IW_IOCTL(SIOCGIWESSID,		aerial_iw_giwessid),
	IW_IOCTL(SIOCSIWNICKN,		NULL),
	IW_IOCTL(SIOCGIWNICKN,		NULL),
	/* Other parameters useful in 802.11 and some other devices */
	IW_IOCTL(SIOCSIWRATE,		aerial_iw_siwrate),
	IW_IOCTL(SIOCGIWRATE,		aerial_iw_giwrate),
	IW_IOCTL(SIOCSIWRTS,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWRTS,		aerial_iw_not_support),
	IW_IOCTL(SIOCSIWFRAG,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWFRAG,		aerial_iw_not_support),
	IW_IOCTL(SIOCSIWTXPOW,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWTXPOW,		aerial_iw_not_support),
	IW_IOCTL(SIOCSIWRETRY,		aerial_iw_not_support),
	IW_IOCTL(SIOCGIWRETRY,		aerial_iw_not_support),
	/* Encoding stuff (scrambling, hardware security, WEP...) */
	IW_IOCTL(SIOCSIWENCODE,		aerial_iw_siwenc),
	IW_IOCTL(SIOCGIWENCODE,		aerial_iw_giwenc),
	/* Power saving stuff (power management, unicast and multicast) */
	IW_IOCTL(SIOCSIWPOWER,		aerial_iw_siwpower),
	IW_IOCTL(SIOCGIWPOWER,		aerial_iw_giwpower),
	/* WPA : Generic IEEE 802.11 informatiom element */
	IW_IOCTL(SIOCSIWGENIE,		NULL),
	IW_IOCTL(SIOCGIWGENIE,		NULL),
	/* WPA : IEEE 802.11 MLME requests */
	IW_IOCTL(SIOCSIWMLME,		NULL),
	/* WPA : Authentication mode parameters */
	IW_IOCTL(SIOCSIWAUTH,		aerial_iw_siwauth),
	IW_IOCTL(SIOCGIWAUTH,		NULL),
	/* WPA : Extended version of encoding configuration */
	IW_IOCTL(SIOCSIWENCODEEXT,	aerial_iw_siwencodeext),
	IW_IOCTL(SIOCGIWENCODEEXT,	NULL),
	/* WPA2 : PMKSA cache management */
	IW_IOCTL(SIOCSIWPMKSA,		NULL),
};

#define IW_PRIV(x, func) [(x) - SIOCIWFIRSTPRIV] = (iw_handler)func
static const iw_handler aerial_iw_priv_handlers[] = {
	IW_PRIV(AERIAL_PRIV_FWLOAD,	aerial_ip_fwload),
	IW_PRIV(AERIAL_PRIV_SETMAC,	aerial_ip_setmac),
	IW_PRIV(AERIAL_PRIV_GETMAC,	aerial_ip_getmac),
	IW_PRIV(AERIAL_PRIV_SETCRYPT,	aerial_ip_setcrypt),
	IW_PRIV(AERIAL_PRIV_GETCRYPT,	aerial_ip_getcrypt),
	IW_PRIV(AERIAL_PRIV_SETPSK,	aerial_ip_setpsk),
	IW_PRIV(AERIAL_PRIV_GETPSK,	aerial_ip_getpsk),
	IW_PRIV(AERIAL_PRIV_SETPSAVE,	aerial_ip_setpsave),
	IW_PRIV(AERIAL_PRIV_GETPSAVE,	aerial_ip_getpsave),
	IW_PRIV(AERIAL_PRIV_SETSLEEP,	aerial_ip_setsleep),
	IW_PRIV(AERIAL_PRIV_GETSLEEP,	aerial_ip_getsleep),

	IW_PRIV(AERIAL_PRIV_RESET,	aerial_ip_reset),

	IW_PRIV(AERIAL_PRIV_SETLOG,	aerial_ip_setlog),
	IW_PRIV(AERIAL_PRIV_GETLOG,	aerial_ip_getlog),
};

#define IW_PRIV_ARGS(_cmd, _name, _set, _get) \
{                                             \
	.cmd = _cmd,                          \
	.name = _name,                        \
	.set_args = _set,                     \
	.get_args = _get,                     \
}
#define IW_ARG_CHAR_T(x) (IW_PRIV_TYPE_CHAR | (x))
#define IW_ARG_INT_FIXED_T(x) (IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | (x))

static const struct iw_priv_args aerial_iw_priv_args[] = {
	IW_PRIV_ARGS(AERIAL_PRIV_FWLOAD, "fwload",
		     0,
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_SETMAC, "set_macaddr",
		     IW_ARG_CHAR_T(12),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETMAC, "get_macaddr",
		     0,
		     IW_ARG_CHAR_T(12)),
	IW_PRIV_ARGS(AERIAL_PRIV_SETCRYPT, "set_cryptmode",
		     IW_ARG_CHAR_T(32),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETCRYPT, "get_cryptmode",
		     0,
		     IW_ARG_CHAR_T(32)),
	IW_PRIV_ARGS(AERIAL_PRIV_SETPSK, "set_psk",
		     IW_ARG_CHAR_T(128),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETPSK, "get_psk",
		     0,
		     IW_ARG_CHAR_T(128)),
	IW_PRIV_ARGS(AERIAL_PRIV_SETPSAVE, "set_power_save",
		     IW_ARG_INT_FIXED_T(1),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETPSAVE, "get_power_save",
		     0,
		     IW_ARG_INT_FIXED_T(1)),
	IW_PRIV_ARGS(AERIAL_PRIV_SETSLEEP, "set_sleep",
		     IW_ARG_INT_FIXED_T(1),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETSLEEP, "get_sleep",
		     0,
		     IW_ARG_INT_FIXED_T(1)),

	IW_PRIV_ARGS(AERIAL_PRIV_RESET, "reset",
		     0,
		     0),

	IW_PRIV_ARGS(AERIAL_PRIV_SETLOG, "set_log_level",
		     IW_ARG_INT_FIXED_T(1),
		     0),
	IW_PRIV_ARGS(AERIAL_PRIV_GETLOG, "get_log_level",
		     0,
		     IW_ARG_INT_FIXED_T(1)),
};

static struct iw_handler_def aerial_iw_handler_def = {
	.standard = (iw_handler *)aerial_iw_handlers,
	.num_standard = ARRAY_SIZE(aerial_iw_handlers),
	.private = (iw_handler *)aerial_iw_priv_handlers,
	.num_private = ARRAY_SIZE(aerial_iw_priv_handlers),
	.private_args = (struct iw_priv_args *)aerial_iw_priv_args,
	.num_private_args = ARRAY_SIZE(aerial_iw_priv_args),
	.get_wireless_stats = aerial_iw_getstats,
};

void
aerial_wireless_attach(struct net_device *dev)
{
	dev->wireless_handlers = &aerial_iw_handler_def;

	return;
}
